/* 
 * Copyright (c) 2010, Lorenz Moesenlechner <moesenle@cs.tum.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sstream>
#include <iterator>
#include <algorithm>

#include <boost/lexical_cast.hpp>

#include <json_prolog/prolog_value.h>

namespace json_prolog
{

std::ostream &operator<<(std::ostream &strm, const PrologValue &v)
{
  return (strm << v.toString());
}

std::string PrologValue::toString() const
{
  switch(type_)
  {
    case DOUBLE:
      return boost::lexical_cast<std::string>(as<double>());
    case INT:
      return boost::lexical_cast<std::string>(as<int64_t>());
    case STRING:
      return as<std::string>();
    case TERM:
    {
      const PrologTerm &trm = as<PrologTerm>();
      std::ostringstream strm;
      strm << trm.name() << "(";
      std::copy(trm.values().begin(), trm.values().end(),
        std::ostream_iterator<PrologValue>(strm, ", "));
      return strm.str().substr(0, strm.str().size()-2) + ")";
    }
    case LIST:
    {
      std::ostringstream strm;
      std::vector<PrologValue> values = as<std::vector<PrologValue> >();
      std::copy(values.begin(), values.end(),
        std::ostream_iterator<PrologValue>(strm, ", "));
      return "[" + strm.str().substr(0, strm.str().size()-2) + "]";
    }
    default:
      return "";
  }
}

}
