import sys
sys.path.insert(0, "/usr/wiss/jain/work/code/SRLDB/datagen")
from datagen import *
from random import choice
from string import *

def decide(p):
    return Decider(p).decide()
          
#Objects in the world
          
class ObjectClass(Object):
    # parent class for all objects (person, utensil, good, place) in the world.
    def __init__(self,world,type,hasFixedConstantName,objectT):
        self.type = type
        self.world = world
        # add a container if necessary
        if world.getContainer(self.type) == None:
            world.addContainer(self.type)
        # initialize the object and add it to the container
        if hasFixedConstantName:
            Object.__init__(self,objectT,capitalize(self.type))
        else:
            Object.__init__(self,objectT)
        world.getContainer(self.type).add(self)
        # set the type attribute
        self[objectT+'T'] = self.type
            
class Place(ObjectClass):
    def __init__(self,world,type,hasFixedConstantName = False):
        ObjectClass.__init__(self,world,type,hasFixedConstantName,'place')
        
class Seat(ObjectClass):
    def __init__(self,world,type,hasFixedConstantName = False):
        #Seat Type should be in the Form Seat##, 
        ObjectClass.__init__(self,world,type,hasFixedConstantName,'seat')
    def toInt(self):
        #this function returns the number of the seat
        return(int(self.type[4:]))
        
class Utensil(ObjectClass):
    def __init__(self,world,type,hasFixedConstantName = False):
        ObjectClass.__init__(self,world,type,hasFixedConstantName,'utensil')
    def setPlace(self,place):
        # sets the current place for this utensil. Note that this doesn't create a relation in the database. 
        self.place = place
    def getPlace(self):
        # returns the current place of the utensil
        return(self.place)
    def setState(self,state):
        self.state = state
    def getState(self):
        return(self.state)
        
class Good(ObjectClass):
    def __init__(self,world,type,hasFixedConstantName = False):
        ObjectClass.__init__(self,world,type,hasFixedConstantName,'good')
    def setPlace(self,place):
        # sets the current place for this good. Note that this doesn't create a relation in the database. 
        self.place = place
    def getPlace(self):
        # returns the current place of the good
        return(self.place)
        
class Person(ObjectClass):
    def __init__(self,world,type,hasFixedConstantName = False):
        ObjectClass.__init__(self,world,type,hasFixedConstantName,'person')
    
    def getName(self):
        return self.type
    
    def setTimeDist(self,timeDist):
        # distribution over the persons time preferences for meals
        self.timeDist = timeDist
        
    def returnPreferredTime(self,mealT,weekend):
        return(AttrDist(self.timeDist[mealT][int(weekend)]).generate())
    
    def setFoodDist(self,foodDist):
        # distribution over the persons consumption preferences for meals
        # a mapping from names of meal types to a 
        # foodDist should be in the form to a list containing two elements, a distribution for drinks and a distribution for food, where each is a mapping from food/drink types to probability values
        # {"Breakfast": [ {distribution for drinks}
        #                      {distribution for food}],
        #   "Lunch": [...],
        #   "CoffeeBreak": [...],
        #   "Dinner": [...]}
        # TODO: How should distribution look like??
        self.foodDist = foodDist
        
    def sampleMeal(self,mealType,usesFor = False):
        # samples the consumption of goods and the usage of utensils of a person in a given meal
        self.uses = []
        self.consumes = []
        self.usesfor = {}
        #if the 'usesForIn' predicate should not generated, we are not interested in a mapping from utensils to goods. Therefore we only return the first two returns from the meal sampling
        if usesFor:
            i = 3
        else:
            i = 2
        if mealType == 'Breakfast': return(self.sampleBreakfast()[:i])
        elif mealType == 'Lunch': return(self.sampleLunch()[:i])
        elif mealType == 'CoffeeBreak': return(self.sampleCB()[:i])
        elif mealType == 'Dinner': return(self.sampleDinner()[:i])
        
    def sampleBreakfast(self):
        self.breakfastFD = self.foodDist['Breakfast']
        dD = self.breakfastFD[0]
        fD = self.breakfastFD[1]
        # samples a meal the person takes part in. Returns all utensils the person uses and all goods the person consumes
        if decide(dD['warm']):
            self.uses.append('cup')
            self.usesfor['cup'] = []
            if dD["coffee"] != 'else' and decide(dD["coffee"]):
                self.consumes.append('coffee')
                self.usesfor['cup'].append('coffee')
            elif dD["coffee"] != "else" and dD["tea"] == "else":
                self.consumes.append('tea')
                self.usesfor['cup'].append('tea')
            if dD["tea"] != 'else' and decide(dD["tea"]):
                self.consumes.append('tea')
                self.usesfor['cup'].append('tea')
            elif dD["tea"] != 'else' and dD["coffee"] == 'else':
                self.consumes.append('coffee')
                self.usesfor['cup'].append('coffee')
        if decide(dD['cold']):
            self.uses.append('glass')
            self.usesfor['glass'] = []
            if dD["juice"] != 'else' and decide(dD["juice"]):
                self.consumes.append('juice')
                self.usesfor['glass'].append('juice')
            elif dD["juice"] != "else" and dD["water"] == "else":
                self.consumes.append('water')
                self.usesfor['glass'].append('water')
            if dD["water"] != 'else' and decide(dD["water"]):
                self.consumes.append('water')
                self.usesfor['glass'].append('water')
            elif dD["water"] != 'else' and dD["juice"] == 'else':
                self.consumes.append('juice')
                self.usesfor['glass'].append('juice')
        if fD["bread"] != 'else' and decide(fD["bread"]):
            self.uses.append('plate')
            self.uses.append('knife')
            self.consumes.append('bread')
            self.usesfor['plate'] = ['bread']
            self.usesfor['knife'] = ['bread']
            for good in ['cheese','sausage']:
                if good in fD and decide(fD[good]): 
                    self.consumes.append(good)
                    self.usesfor['plate'].append(good)
                    self.usesfor['knife'].append(good)
        elif fD["bread"] != "else" and fD["cereals"] == "else":
            self.uses.append('bowl')
            self.uses.append('spoon')
            self.consumes.append('cereals')
            self.usesfor['bowl'] = ['cereals']
            self.usesfor['spoon'] = ['cereals']
        if fD["cereals"] != 'else' and decide(fD["cereals"]):
            self.uses.append('bowl')
            self.uses.append('spoon')
            self.consumes.append('cereals')
            self.usesfor['bowl'] = ['cereals']
            self.usesfor['spoon'] = ['cereals']
        elif fD["cereals"] != 'else' and fD["bread"] == 'else':
            self.uses.append('plate')
            self.uses.append('knife')
            self.consumes.append('bread')
            self.usesfor['plate'] = ['bread']
            self.usesfor['knife'] = ['bread']
            for good in ['cheese','sausage']:
                if good in fD and decide(fD[good]): 
                    self.consumes.append(good)
                    self.usesfor['plate'].append(good)
                    self.usesfor['knife'].append(good)
        return(self.uses, self.consumes, self.usesfor)
    
    def sampleLunch(self):
        self.lunchFD = self.foodDist['Lunch']
        dD = self.lunchFD[0]
        fD = self.lunchFD[1]
        if decide(dD['cold']):
            self.uses.append('glass')
            self.usesfor['glass'] = []
            if dD["juice"] != 'else' and decide(dD["juice"]):
                self.consumes.append('juice')
                self.usesfor['glass'].append('juice')
            elif dD["juice"] != "else" and dD["water"] == "else":
                self.consumes.append('water')
                self.usesfor['glass'].append('water')
            if dD["water"] != 'else' and decide(dD["water"]):
                self.consumes.append('water')
                self.usesfor['glass'].append('water')
            elif dD["water"] != 'else' and dD["juice"] == 'else':
                self.consumes.append('juice')
                self.usesfor['glass'].append('juice')
        self.uses.append('plate')
        self.usesfor['plate'] = []
        self.usesfor['knife'] = []
        self.usesfor['fork'] = []
        self.usesfor['bowl'] = []
        self.usesfor['spoon'] = []
        if decide(fD['pizza']):
            self.consumes.append('pizza')
            self.uses.append('knife')
            self.usesfor['plate'].append('pizza')
            self.usesfor['knife'].append('pizza')
            if decide(0.8): 
                self.uses.append('fork')
                self.usesfor['fork'].append('pizza')
        else:
            self.consumes.append('noodles')
            self.uses.append('fork')
            self.uses.append('spoon')
            self.usesfor['fork'].append('noodles')
            self.usesfor['spoon'].append('noodles')
            self.usesfor['plate'].append('noodles')
        if fD["salad"] != 'else' and decide(fD["salad"]):
            if 'plate' not in self.uses: self.uses.append('plate')
            if 'fork' not in self.uses: self.uses.append('fork')
            self.consumes.append('salad')
            self.usesfor['plate'].append('salad')
            self.usesfor['fork'].append('salad')
        elif fD["salad"] != "else" and fD["soup"] == "else":
            self.uses.append('bowl')
            if 'spoon' not in self.uses: self.uses.append('spoon')
            self.consumes.append('soup')
            self.usesfor['bowl'].append('soup')
            self.usesfor['spoon'].append('soup')
        if fD["soup"] != 'else' and decide(fD["soup"]):
            self.uses.append('bowl')
            if 'spoon' not in self.uses: self.uses.append('spoon')
            self.consumes.append('soup')
            self.usesfor['bowl'].append('soup')
            self.usesfor['spoon'].append('soup')
        elif fD["soup"] != 'else' and fD["salad"] == 'else':
            if 'plate' not in self.uses: self.uses.append('plate')
            if 'fork' not in self.uses: self.uses.append('fork')
            self.consumes.append('salad')
            self.usesfor['plate'].append('salad')
            self.usesfor['fork'].append('salad')
        return(self.uses, self.consumes, self.usesfor)
    def sampleCB(self):
        self.CBFD = self.foodDist['CoffeeBreak']
        dD = self.CBFD[0]
        fD = self.CBFD[1]
        self.uses.append('cup')
        self.usesfor['cup'] = []
        if dD["coffee"] != 'else' and decide(dD["coffee"]):
            self.consumes.append('coffee')
            self.usesfor['cup'].append('coffee')
        elif dD["coffee"] != "else" and dD["tea"] == "else":
            self.consumes.append('tea')
            self.usesfor['cup'].append('tea')
        if dD["tea"] != 'else' and decide(dD["tea"]):
            self.consumes.append('tea')
            self.usesfor['cup'].append('tea')
        elif dD["tea"] != 'else' and dD["coffee"] == 'else':
            self.consumes.append('coffee')
            self.usesfor['cup'].append('coffee')
        self.uses.append('plate')
        self.usesfor['plate'] = []
        if fD["cake"] != 'else' and decide(fD["cake"]):
            self.consumes.append('cake')
            self.usesfor['plate'].append('cake')
        elif fD["cake"] != "else" and fD["bread"] == "else":
            self.consumes.append('bread')
            self.usesfor['plate'].append('bread')
        if fD["bread"] != 'else' and decide(fD["bread"]):
            self.consumes.append('bread')
            self.usesfor['plate'].append('bread')
        elif fD["bread"] != 'else' and fD["cake"] == 'else':
            self.consumes.append('cake')
            self.usesfor['plate'].append('cake')
        return(self.uses, self.consumes, self.usesfor)
    def sampleDinner(self):
        self.dinnerFD = self.foodDist['Dinner']
        dD = self.dinnerFD[0]
        fD = self.dinnerFD[1]
        if decide(dD['warm']):
            self.uses.append('cup')
            self.consumes.append('tea')
            self.usesfor['cup'] = ['tea']
        if decide(dD['cold']):
            self.uses.append('glass')
            self.usesfor['glass'] = []
            if dD["juice"] != 'else' and decide(dD["juice"]):
                self.consumes.append('juice')
                self.usesfor['glass'].append('juice')
            elif dD["juice"] != "else" and dD["water"] == "else":
                self.consumes.append('water')
                self.usesfor['glass'].append('water')
            if dD["water"] != 'else' and decide(dD["water"]):
                self.consumes.append('water')
                self.usesfor['glass'].append('water')
            elif dD["water"] != 'else' and dD["juice"] == 'else':
                self.consumes.append('juice')
                self.usesfor['glass'].append('juice')
        self.uses.append('plate')
        good = AttrDist(fD).generate()
        self.consumes.append(good)
        if good == 'bread':
            self.uses.append('knife')
            self.usesfor['plate'] = ['bread']
            self.usesfor['knife'] = ['bread']
        elif good == 'noodles':
            self.uses.append('fork')
            self.uses.append('spoon')
            self.usesfor['plate'] = ['noodles']
            self.usesfor['fork'] = ['noodles']
            self.usesfor['spoon'] = ['noodles']
        elif good == 'pizza':
            self.uses.append('knife')
            self.usesfor['plate'] = ['pizza']
            self.usesfor['knife'] = ['pizza']
            if decide(0.8): 
                self.uses.append('fork')
                self.usesfor['fork'] = ['pizza']
        return(self.uses, self.consumes, self.usesfor)
       
      
# family of persons in the world.
class Family:
    def __init__(self,name,members = []):
        self.members = members
        self.name = name
    def addMember(self,person):
        # adds a person from the world to the family
        self.members.append(person)
    def sampleNumItemsNeeded(self,mealType,utensils,goods):
        # samples a meal where all members of the family take part in
        self.utensilsNeeded = {}
        self.goodsNeeded = {}
        for utensilT in utensils:
            self.utensilsNeeded[utensilT] = 0
        for goodT in goods:
            self.goodsNeeded[goodT] = 0
        for person in self.members:
            # required utensils and food are sampled for each member of the family
            ut, fd = person.sampleMeal(mealType)
            for utensilT in ut:
                self.utensilsNeeded[utensilT] += 1
            for goodT in fd:
                self.goodsNeeded[goodT] += 1
        return self.utensilsNeeded, self.goodsNeeded
              
# classes for person objects:
def Anna(world, personClass = Person, hasFixedConstantName = False):
    self = personClass(world,'Anna',hasFixedConstantName)
    timeDist = {'Breakfast': [{"earlyMorning": 0.9, "lateMorning": 0.1, "noon": 0.0},{"earlyMorning": 0.8, "lateMorning": 0.2, "noon": 0.0}],
                        'Lunch': [{"noon": 0.9, "earlyAfternoon": 0.1},{"noon": 0.8, "earlyAfternoon": 0.2}],
                        'Dinner': [{'lateAfternoon': 0.6, 'evening': 0.4},{'lateAfternoon': 0.5, 'evening': 0.5}]}
    self.setTimeDist(timeDist)
    foodDist = {"Breakfast": [{"warm": 1.0, "cold": 0.1,"coffee": 0.9, "tea": "else", 'juice': 1.0, 'water': 0.0},
                                            {"bread": 0.95, "cereals": "else", "cheese": 1}],
                "Lunch": [{"cold": 0.8, "water": 0.7, 'juice': 'else'},
                        {"pizza": 0.8, 'salad': 0.9, 'soup': 0.3}],
                "CoffeeBreak": [{'coffee': 0.9, 'tea': 'else'},
                        {'cake' :1.0, 'bread': 0.0}],
                "Dinner": [{"warm": 0.2, "cold": 0.9, 'water': 0.7, 'juice': 'else'},
                        {'bread': 0.9, 'noodles': 0.05, 'pizza': 0.05}]}
    self.setFoodDist(foodDist)
    return self
        
def Bert(world, personClass = Person, hasFixedConstantName = False):
    self = personClass(world,'Bert',hasFixedConstantName)
    timeDist = {'Breakfast': [{"earlyMorning": 0.9, "lateMorning": 0.1, "noon": 0.0},{"earlyMorning": 0.8, "lateMorning": 0.2, "noon": 0.0}],
                        'Lunch': [{"noon": 0.9, "earlyAfternoon": 0.1},{"noon": 0.8, "earlyAfternoon": 0.2}],
                        'Dinner': [{'lateAfternoon': 0.6, 'evening': 0.4},{'lateAfternoon': 0.5, 'evening': 0.5}]}
    self.setTimeDist(timeDist)
    foodDist = {"Breakfast": [{"warm": 1.0, "cold": 0.8, "coffee": "else", "tea": 0.9, 'juice': 0.9, 'water': 'else'},
                            {"bread": 0.95, "cereals": "else", "cheese": 0.9, "sausage": 0.9}],
                "Lunch": [{"cold": 0.7, "water": 'else', 'juice': 0.9},
                        {"pizza": 0.6, 'salad': 0.4, 'soup': 0.2}],
                "CoffeeBreak": [{'coffee': 'else', 'tea': 0.9},
                        {'cake' :0.9, 'bread': 'else'}],
                "Dinner": [{"warm": 'else', "cold": 0.7, 'water': 'else', 'juice': 0.9},
                        {'bread': 0.8, 'noodles': 0.2, 'pizza': 0.0}]}
    self.setFoodDist(foodDist)
    return self
            
def Charly(world, personClass = Person, hasFixedConstantName = False):
    self = personClass(world,'Charly',hasFixedConstantName)
    timeDist = {'Breakfast': [{"earlyMorning": 0.5, "lateMorning": 0.5, "noon": 0.0},{"earlyMorning": 0.2, "lateMorning": 0.8, "noon": 0.0}],
                        'Lunch': [{"noon": 0.5, "earlyAfternoon": 0.5},{"noon": 0.2, "earlyAfternoon": 0.8}],
                        'Dinner': [{'lateAfternoon': 0.7, 'evening': 0.3},{'lateAfternoon': 0.4, 'evening': 0.6}]}
    self.setTimeDist(timeDist)
    foodDist = {'Breakfast': [{"warm": 1.0, "cold": 0.2, "coffee": "else", "tea": 0.9, 'juice': 0.0, 'water': 1.0},
                            {"bread": "else", "cereals": 0.95, "cheese": 0.9, "sausage": 0.9}],
                "Lunch": [{"cold": 0.4, "water": 1.0, 'juice': 0.0},
                        {"pizza": 0.2, 'salad': 0.9, 'soup': 'else'}],
                "CoffeeBreak": [{'coffee': 'else', 'tea': 0.9},
                        {'cake' :0.7, 'bread': 'else'}],
                "Dinner": [{"warm": 0.8, "cold": 0.9, 'water': 1.0, 'juice': 0.0},
                        {'bread': 0.6, 'noodles': 0.3, 'pizza': 0.1}]}
    self.setFoodDist(foodDist)
    return self
            
def Dorothy(world, personClass = Person, hasFixedConstantName = False):
        self = personClass(world,'Dorothy',hasFixedConstantName)
        timeDist = {'Breakfast': [{"earlyMorning": 0.2, "lateMorning": 0.8, "noon": 0.0}, {"earlyMorning": 0.2, "lateMorning": 0.8, "noon": 0.0}],
                            'Lunch': [{"noon": 0.2, "earlyAfternoon": 0.8},{"noon": 0.2, "earlyAfternoon": 0.8}],
                            'Dinner': [{'lateAfternoon': 0.3, 'evening': 0.7},{'lateAfternoon': 0.2, 'evening': 0.8}]}
        self.setTimeDist(timeDist)
        foodDist = {'Breakfast': [{"warm": 0.2, "cold": 0.9, "coffee": "else", "tea": 0.9, 'juice': 0.9, 'water': 'else'},
                                {"bread": "else", "cereals": 0.95, "cheese": 0.9, "sausage": 0.0}],
                    "Lunch": [{"cold": 0.9, "water": 'else', 'juice': 0.9},
                            {"pizza": 0.4, 'salad': 'else', 'soup': 0.6}],
                    "CoffeeBreak": [{'coffee': 'else', 'tea': 0.9},
                            {'cake' :0.6, 'bread': 'else'}],
                    "Dinner": [{"warm": 0.2, "cold": 0.9, 'water': 'else', 'juice': 0.9},
                            {'bread': 0.4, 'noodles': 0.3, 'pizza': 0.3}]}
        self.setFoodDist(foodDist)
        return self
        
def Emily(world, personClass = Person, hasFixedConstantName = False):
        self = personClass(world,'Emily',hasFixedConstantName)
        timeDist = {'Breakfast': [{"earlyMorning": 0.1, "lateMorning": 0.9, "noon": 0.0},{"earlyMorning": 0.0, "lateMorning": 0.8, "noon": 0.2}],
                            'Lunch': [{"noon": 0.1, "earlyAfternoon": 0.9},{"noon": 0.0, "earlyAfternoon": 1.0}],
                            'Dinner': [{'lateAfternoon': 0.2, 'evening': 0.8},{'lateAfternoon': 0.1, 'evening': 0.9}]}
        self.setTimeDist(timeDist)
        foodDist = {'Breakfast': [{"warm": 0.9, "cold": 0.9, "coffee": 0.9, "tea": 'else', 'juice': 0.9, 'water': 'else'},
                                {"bread": 0.8, "cereals": 0.8, "cheese": 0.9, "sausage": 0.9}],
                    "Lunch": [{"cold": 0.9, "water": 'else', 'juice': 0.9},
                            {"pizza": 0.7, 'salad': 0.5, 'soup': 0.5}],
                    "CoffeeBreak": [{'coffee': 0.9, 'tea': 'else'},
                            {'cake' :0.8, 'bread': 'else'}],
                    "Dinner": [{"warm": 0.9, "cold": 0.9, 'water': 'else', 'juice': 0.9},
                            {'bread': 0.8, 'noodles': 0.0, 'pizza': 0.2}]}
        self.setFoodDist(foodDist)
        return self
        
def Frank(world, personClass = Person, hasFixedConstantName = False):
        self = personClass(world,'Frank',hasFixedConstantName)
        timeDist = {'Breakfast': [{"earlyMorning": 0.0, "lateMorning": 0.7, "noon": 0.3},{"earlyMorning": 0.0, "lateMorning": 0.3, "noon": 0.7}],
                            'Lunch': [{"noon": 0.1, "earlyAfternoon": 0.9},{"noon": 0.0, "earlyAfternoon": 1.0}],
                            'Dinner': [{'lateAfternoon': 0.3, 'evening': 0.7},{'lateAfternoon': 0.1, 'evening': 0.9}]}
        self.setTimeDist(timeDist)
        foodDist = {'Breakfast': [{"warm": 0.9, "cold": 0.9, "coffee": "else", "tea": 0.9, 'juice': "else", 'water': 0.9},
                                {"bread": 0.9, "cereals": 0.2, "cheese": 0.9, "sausage": 0.9}],
                    "Lunch": [{"cold": 0.9, "water": 0.9, 'juice': 'else'},
                            {"pizza": 0.3, 'salad': 0.7, 'soup': 'else'}],
                    "CoffeeBreak": [{'coffee': 'else', 'tea': 0.9},
                            {'cake' :0.3, 'bread': 'else'}],
                    "Dinner": [{"warm": 0.9, "cold": 0.9, 'water': 0.9, 'juice': 'else'},
                            {'bread': 0.1, 'noodles': 0.1, 'pizza': 0.8}]}
        self.setFoodDist(foodDist)
        return self

if __name__ == '__main__':
    world = World()
    Place('Table')
    Place('Heater')
    world.printAll()
    
