import sys
from datagen import *

class MealGen(ObjectGenerator):
	def __init__(self, world,persons, seats):
		ObjectGenerator.__init__(self, "meal", world.getContainer("event"), {})
		self.world = world
		self.persons = persons
		self.seats = seats

	def _createAttributes(self, obj, **args):
		days = ["Monday","Tuesday","Wednesday","Thursday","Friday","Saturday","Sunday"]
		weekend = args["day"] >= 5
		obj.setAttr("day", days[args["day"]])
		obj.setAttr("weekend", strBool(weekend))
		obj.setAttr("mealT", args["type"])
			
	def _createLinks(self, event, **args):
		world = self.world
		if event['mealT'] == 'Breakfast':
			times = {"earlyMorning": 0.0, "lateMorning": 0.0, "noon": 0.0}
		elif event['mealT'] == 'Lunch':
			times = {"noon": 0.0, "earlyAfternoon": 0.0}
		elif event['mealT'] == 'Dinner':
			times = {"lateAfternoon": 0.0, "evening": 0.0}
			
		print "\n%s" % event['mealT']
		
		# determine people who participate depending on day of week
		if event.getAttr("day") == "Monday":
			participate = ["Anna","Bert"]
			if decide(0.4): participate.append("Charly")
		elif event.getAttr("day") == "Tuesday":
			participate = ["Dorothy","Frank","Anna"]
			if decide(0.8): participate.append("Bert")
		elif event.getAttr("day") == "Wednesday":
			participate = ["Anna","Bert"]
			if decide(0.8): participate.append("Dorothy")
			if decide(0.8): participate.append("Emily")
		elif event.getAttr("day") == "Thursday":
			participate = ["Dorothy","Emily","Charly"]
		elif event.getAttr("day") == "Friday":
			participate = ["Charly","Emily"]
			if decide(0.8): participate.append("Dorothy")
		elif event.getAttr("day") == "Saturday":
			participate = ["Emily","Frank"]
			if decide(0.8): participate.append("Charly")
			if decide(0.8): participate.append("Bert")
		elif event.getAttr("day") == "Sunday":
			participate = ["Frank","Dorothy"]
			if decide(0.8): participate.append("Emily")
		
		# seat selection and consumption and utensil usage
		for person in participate:
			self.persons[person][0].takesPartIn(event,self.persons,self.seats)
			self.persons[person][0].setConsumptionLinks(event,'standard','CAO',True)
			times[self.persons[person][0].time] += 1            
			
		# determine time of meal by building a distribution using the preferred times of the participants
		event["timeT"] = AttrDist(times).generate()
		print "  time: %s, %s  %s" % (event["day"], event["timeT"], str(times))
		
		# determine usage of shared objects
		pitcher = Chooser(world,'pitcher').choose()
		platter = Chooser(world,'platter').choose()
		if decide(0.8): 
			pitcher.linkto("usedByAllIn",event)
			ConstantAsObject(pitcher["utensilT"], self.world).linkto("anyUsedByAllIn", event)
		if decide(0.8): 
			platter.linkto("usedByAllIn",event)
			ConstantAsObject(platter["utensilT"], self.world).linkto("anyUsedByAllIn", event)
		
		world.restoreAllRemovedObjects()


# --- main app ---

from sys import argv
argv = argv[1:]
if len(argv) != 1:
	print "usage: gen <number of weeks>"

numWeeks = int(argv[0])

world = World()

# generate places
seats = {}
for i in range(numSeats):
	seatT = "Seat" +str(i)
	world.addContainer(seatT)
	seats[seatT] = []    
	seats[seatT].append(Seat(world,seatT,True))

# generate kitchen utensils
utensilTypes = ["cup", "glass", "bowl", "plate", "spoon", "fork", "knife","napkin"]
utensils = {}
for utensilT in utensilTypes:
	world.addContainer(utensilT)
	utensils[utensilT] = []
	for i in range(numUtensils):            
		utensils[utensilT].append(Utensil(world,utensilT))

#community utensils
comUtensilTypes = ["pitcher","platter"]
for comUtensilT in comUtensilTypes:
	world.addContainer(comUtensilT)
	utensils[comUtensilT] = []
	utensils[comUtensilT].append(Utensil(world,comUtensilT))
		   
# generate goods
goodsTypes = ["cheese","sausage", "cereals", "bread", "coffee", "tea", "water", "juice", 'pizza', 'noodles', 'salad', 'soup', 'cake', 'milk']
goods = {}
for goodsT in goodsTypes:
	world.addContainer(goodsT)
	goods[goodsT] = []
	goods[goodsT].append(Good(world,goodsT,True))

# generate kitchen users
peopleT = ["Anna","Bert","Charly","Dorothy","Emily","Frank"]
people = {}
i = 1
for personName in peopleT:
	people[personName] = []
	world.addContainer(personName)
	person = eval('%s(world, PersonNEW, peopleAsConstants)' % personName)
	person.setFavoriteSeat(seats['Seat'+str(i%numSeats)][0])
	person["name"] = personName
	people[personName].append(person)    
	i += 1

# generate events    
events = world.addContainer("event")
mealgen = MealGen(world,people,seats)
mealTypes = ['Breakfast','Lunch','Dinner']
for day in range(7*numWeeks):
	print "\n*** DAY %d ***\n" % day+1
	for mT in mealTypes:
		mealgen.generate(day=day % 7, type=mT)

# print all generated objects
world.printAll()

# return the database
db = world.getDatabase()
